// ===================================================================
// $Id: $
//
// CleanMesh.h
//
// Class: CCleanMesh
//
//  The header file to implement rendering via OpenGL with shaders (without texture)
//
// Initial coding by Vlastimil Havran, 2013.

#ifndef __CLEANMESH_H__
#define __CLEANMESH_H__

// use the library for OpenGL
#include "pgr.h"
#include <GL/freeglut.h>
//#include <GL/glext.h>
#include <GL/gl.h>
#include <GL/glu.h>

// in order to get function prototypes from glext.h, define
//#GL_GLEXT_PROTOTYPES before including glext.h
//#define GL_GLEXT_PROTOTYPES

#include "glext.h"
//#include "glInfo.h"

#include "pgr.h"
#include <cstdio>
#include <cstdlib>
#include <cmath>
#include <vector>
#include <cassert>
#include <cstring>
#include <sys/stat.h>
#include <cmath>
#include <iostream>
#include <sstream>
#include <string>
#include <iomanip>

using namespace std;

#include "configg.h"
#include "bconfig.h"

// ----------------------------------------------------
// This class implements textured triangle mesh
class CCleanMesh
{  
  // internal data connected to vertices
  int cntVertices; // number of vertices to be rendered
  float *vertices; // 3 floats per vertex
  float ccolor[3]; // the color to be used

  // Arrays for MESH
  // Vertex buffer object
  GLuint vaoHandle;
  // vertex array object for vertices only
  GLuint vbovertices; // vbonormals;
  GLint vertexLoc; // attribute index
  // Normals
  // GLuint vbonormals;
  // GLint normaLoc; // attribute index

  // initialize shaders
  bool createShaderProgramNoNormals();

  // How to render the data
  GLenum renderMode;
  GLuint shaderProgram; // shader program handle in OpenGL
  void copyVertex(float *dest, float *src, int d, int s) {
    dest[d+0] = src[s+0]; dest[d+1] = src[s+1]; dest[d+2] = src[s+2];
  }
  // rotate individual point
  void rotate3DPoint(float angle, float *v, int axis);
public:
  // constructor
  CCleanMesh(): cntVertices(0), vertices(0), vaoHandle(0), vbovertices(0),
		vertexLoc(0), renderMode(0), shaderProgram(0) { }
  ~CCleanMesh() { free(); }
  void init(int cntVertices, int renderModeV, float *verticesV);
  void setColor(float red, float green, float blue) {
    ccolor[0] = red; ccolor[1] = green; ccolor[2] = blue;
  }
  void free();
  void setRenderMode(int renderModeV);
  void toggleLinesTriangles(int mode); // 1 .. convert to lines, 0 .. convert to triangles from lines
  // draw the mesh to the output
  void draw(const glm::mat4 &matrix);
  // rotate object geometry by an angle around axis
  void rotateOBJModelGeometry(int axis, float rotAngle);
  // translate object geometry by a shift in some axis
  void translateOBJModelGeometry(int axis, float value);
};

#endif // __CLEANMESH_H__
