// ===================================================================
// $Id: $
//
// OBJloader.h
//
//     Header file for Wawefront OBJ file geometry loading
//
// Initial coding by Jiri Filip, 2004.

#ifndef __OBJLOADER_H__
#define __OBJLOADER_H__

#include <cstdio>
#include <cstdlib> 

class TObjectData
{
private:
  int dataSize;
  long i_v, i_vt, i_vn, i_f;
  char *fileName;
  char *OBJfileName;

public:
  float **vertices;
  float **UVcoord;
  float **normals;
  long **faces;
  float **tangentsF;  // tangents for individual faces 
  float **tangentsV;  // tangents for individual vertices
  float **normalsF;   // normals for individual faces
  float **normalsV;   // normals for individual vertices
  float **binormalsV;  // binormals for individual vertices
  float **binormalsF;  // binormals for individual faces

public:
  TObjectData(int dataSize);
  ~TObjectData();
  int getNoOfTokens(char *linem, char *delimiter);
  int loadOBJfile(char *_fileName, int center, float scale, float xoff, float yoff, float zoff);
  void normalsCorrection();
  void saveLoadedOBJ(char *filePrefix, int off_v, int off_vt, int off_vn);
  void splitOBJ(char *fileName);
  long noOfFaces();
  long noOfVertices();
  long noOfNormals();

  void testOnDegeneratedPolygons();
  void testAndCorrectNormalsContinuity();

  void normalsPerFace();
  void normalsPerVertex();

  void computeTB(float **v, float **UV, float *T, float *B);
  void tangentsPerFace();
  void tangentsPerVertex();

  void saveTangentsPerVertex(char *fileName);
  bool loadTangentsPerVertex(char *fileName);

  void binormalsPerFace();
  void binormalsPerVertex();

  //! get "vertex"th vertex coordinate "xyz" of face "face"
  float getV(int face, int vertex, int xyz);

  //! get "vertex"th vertex index of face "face"
  int getVindex(int face, int vertex);

  //! get "vertex"th normal coordinate "xyz" of face "face"
  float getN(int face, int vertex, int xyz);
  //! get "vertex"th interpolated normal coordinate "xyz" of face "face" - 3 normals for the face
  float getNv(int face, int vertex, int xyz);
  //! get normal coordinate "xyz" of face "face" - 1 normal for face 
  float getNf(int face, int xyz);

  //! get "vertex"th interpolated tangent coordinate "xyz" of face "face" - 3 tangents for the face
  float getT(int face, int vertex, int xyz);
  //! get tangent coordinate "xyz" of face "face" - 1 tangent for face 
  float getTf(int face, int xyz);

  //! get "vertex"th interpolated binormal coordinate "xyz" of face "face" - 3 binormals for face
  float getB(int face, int vertex, int xyz);
  //! get binormal coordinate "xyz" of face "face" - 1 binormal for face 
  float getBf(int face, int xyz);

  //! get "vertex"th UV texture coordinate "uv" of face "face" - 3 UVs for face
  float getUV(int face, int vertex, int uv);

  void vec_crossprod(float *x, float *y, float *res);
  float vec_len(float *x);
  void vec_diff(float *x, float *y, float *res);
  int vec_norm(float *x);
  void normal(float *v1, float *v2, float *v3, float *n);

};

#endif // __OBJLOADER_H__

