// ===================================================================
// $Id: $
//
// TexturedMesh.h
//
// Class: CTexturedMesh
//
//  The class to implement rendering via OpenGL with shaders and HDR texture
//
// Initial coding by Vlastimil Havran, 2013.

#ifndef __TEXTURED_MESH_H__
#define __TEXTURED_MESH_H__

// use the library for OpenGL
#include "pgr.h"
#include <GL/freeglut.h>
//#include <GL/glext.h>
#include <GL/gl.h>
#include <GL/glu.h>

// in order to get function prototypes from glext.h, define
//#GL_GLEXT_PROTOTYPES before including glext.h
//#define GL_GLEXT_PROTOTYPES

#include "glext.h"
//#include "glInfo.h"

#include "pgr.h"
#include <cstdio>
#include <cstdlib>
#include <cmath>
#include <vector>
#include <cassert>
#include <cstring>
#include <sys/stat.h>
#include <cmath>
#include <iostream>
#include <iostream>
#include <sstream>
#include <string>
#include <iomanip>

using namespace std;

#include "configg.h"
#include "bconfig.h"

// ----------------------------------------------------
// This class implements textured triangle mesh
class CTexturedMesh
{  
  // internal data connected to vertices
  int cntVertices; // number of vertices to be rendered
  float *vertices; // 3 floats per vertex
  float *texcoords; // 2 floats per vertex
  float *texturedata; // the data as texture
  //float *normals; // 3 floats per vertex
  //float *tangents; // 3 floats per vertex
  //float *bitangents; // 3 floats per vertex

  // Arrays for MESH
  // Vertex buffer object
  GLuint vaoHandle;
  // vertex array object for vertices and UV texture coordinates
  GLuint vbovertices, vbotexcoords; // vbonormals;
  GLint vertexLoc; // attribute index

  // Texture object
  GLuint texobject;
  // texture size
  int tex_width, tex_height;  
  GLint texLoc; // attribute index
  int idTextureUnit; // which texture unit should be used
  // Normals
  // GLuint vbonormals;
  // GLint normaLoc; // attribute index

  // initialize shaders
  bool createShaderProgramNoNormals();

  // How to render the data
  GLenum renderMode;
  GLuint shaderProgram; // shader program handle in OpenGL
  void copyVertex(float *dest, float *src, int d, int s) {
    dest[d+0] = src[s+0]; dest[d+1] = src[s+1]; dest[d+2] = src[s+2];
  }
  // rotate individual point
  void rotate3DPoint(float angle, float *v, int axis);
public:
  // constructor
  CTexturedMesh(): cntVertices(0), vertices(0), texcoords(0), texturedata(0),
		   vaoHandle(0), vbovertices(0), vbotexcoords(0), vertexLoc(0),
		   texobject(0), tex_width(0), tex_height(0), texLoc(0),
		   renderMode(0), shaderProgram(0) { }
  ~CTexturedMesh() { free(); }
  void init(int cntVertices, int renderModeV, float *verticesV, float *texcoordsV, int idTextureUnit=0);
  int getCntVertices() const { return cntVertices;}
  void free();
  // If lines, triangles, points etc. as defined by OpenGL
  void setRenderMode(int renderModeV);
  void toggleLinesTriangles(int mode); // 1 .. convert to lines, 0 .. convert to triangles from lines
  // load the files and creates texture with all attributes
  void loadTexture(const char *filename);
  // expects floating point texture in RGB format
  void setTextureData(float *pointer, int tex_width, int tex_height);
  void detachTextureData();
  // draw the mesh to the output
  void draw(const glm::mat4 &matrix);
  // do not bind the texture and assume some texture is already bound
  //void drawNBT(const glm::mat4 &matrix);
  // rotate object geometry by an angle around axis
  void rotateOBJModelGeometry(int axis, float rotAngle);
  // translate object geometry by a shift in some axis
  void translateOBJModelGeometry(int axis, float value);
};

#endif // __TEXTURED_MESH_H__
