// ===================================================================
// $Id: $
//
// glenvmap.cpp
//
// Class: CEnvironmentMapGL
//
//  The source file of the class for environment map via OpenGL using shaders
//
// Initial coding by Vlastimil Havran, 2013.

#include "pgr.h"
#include <GL/freeglut.h>
//#include <GL/glext.h>
#include <GL/gl.h>
#include <GL/glu.h>

// in order to get function prototypes from glext.h, define
//#GL_GLEXT_PROTOTYPES before including glext.h
//#define GL_GLEXT_PROTOTYPES

#include "glext.h"
//#include "glInfo.h"

#include "pgr.h"
#include <cstdio>
#include <cstdlib>
#include <cmath>
#include <vector>
#include <cassert>
#include <cstring>
#include <sys/stat.h>
#include <cmath>
#include <iostream>
#include <iostream>
#include <sstream>
#include <string>
#include <iomanip>

using namespace std;

#include "configg.h"
#include "bconfig.h"
#include "basstr.h"
#include "hdrloader.h"
#include "glenvmap.h"


static std::string vertexShader2 =
"#version 140\n"
"uniform mat4 PVMmatrix;\n"
"in vec3 vertex;\n"
"out vec4 position_v;\n"
"smooth out vec3 texcords3;\n"
"void main() {\n"
"  position_v = PVMmatrix * vec4(vertex, 1);\n"
"  gl_Position = position_v;\n"
"  texcords3 = vertex;\n"
"}\n"
;
// "  //position_v = vec4(vertex, 1);\n"

static std::string fragmentShader =
"#version 140\n"
"uniform mat4 Mmatrix;\n"
"uniform sampler2D texsampler;\n"
"in vec4 position_v;\n"
"in vec3 texcords3;\n"
"out vec4 outputColor;\n"
"void main()\n"
"{\n"
"  vec3 tt = normalize(texcords3);\n"
"  vec2 texcoords;\n"
"  texcoords.x = fract(((atan(tt.y, tt.x)+3.1415927f)/6.283f)+0.5f);\n"
"  texcoords.y = fract(acos(tt.z)/3.1415927f);\n"
"  outputColor = texture(texsampler, texcoords);\n"
"}\n"
;
// "  vec3 tt = normalize(vec3(position_v));\n"
// "  vec3 tt = normalize(vec3(Mmatrix * position_v));\n"
//"  if ((texcoords.y > 0.1) && (texcoords.y < 0.15))\n"
//"    outputColor = vec4(1, 0, 0, 0);\n"
//"  else\n"
//"    outputColor = texture(texsampler, texcoords);\n"

// initialize shaders
bool
CEnvironmentMapGL::createShaderProgramNoNormals()
{
  // load shader
  std::vector<GLuint> shaders;
  // load vertex shader
  int glIDvs = pgr::createShaderFromSource(GL_VERTEX_SHADER, vertexShader2);
  if (glIDvs > 0)
    shaders.push_back(glIDvs);
  else {
    cout << "ERROR" << endl;
    return true; // error
  }
  // load fragment shader
  int glIDfs = pgr::createShaderFromSource(GL_FRAGMENT_SHADER, fragmentShader);
  if (glIDfs > 0)
    shaders.push_back(glIDfs);
  else {
    cout << "ERROR" << endl;
    return true; // error
  }

  // generate program to run
  envShaderProgram = pgr::createProgram(shaders);
  cout << "Init envShaderProgram = " << envShaderProgram << endl;
  //assert(envShaderProgram > -1);
  return false; // OK
}

void
CEnvironmentMapGL::setRenderMode(int renderModeV)
{
  if (!((renderModeV == GL_POINTS)||
	(renderModeV == GL_LINES) ||
	(renderModeV == GL_TRIANGLES))) {
    cout << "Wrong rendering mode = " << renderModeV << endl;
    exit(-1);
  }
  // copy render mode
  renderMode = renderModeV;
}

// Prepare the list of triangles to be rendered
int
CEnvironmentMapGL::generateEnv(float radius, int cntTriangles)
{
  unsigned int vertIdx = 0;
  int THETA_N = sqrt(cntTriangles/4);
  int PHI_N = THETA_N*2;
  int cntVertices = THETA_N*PHI_N*2*3; // the number of vertices
  vertices = new float[cntVertices*3]; // allocate array, each vertex 3 floats
  assert(vertices);
  //cout << "cntTriangles = " << cntTriangles << " cntVertices = " << cntVertices << endl;
  
  // Generate triangles on environment map, + z-axis in the direction of norh pole

  //float dz = 2.0/(float)THETA_N;
  for(int theta=0; theta<THETA_N; theta++) {
    float vz = -1.0f + 2.0f*(float)theta/(float)THETA_N;
    float vz2 = -1.0f + 2.0f*(float)(theta+1)/(float)THETA_N;
    for(int phi=0; phi<PHI_N; phi++) {
      float phiv = 2.0f * M_PI * phi/(float)PHI_N;
      float phiv2 = 2.0f * M_PI * (phi+1)/(float)PHI_N;
      // The first triangle with 3 vertices
      // First vertex coordinates
      float radxy = radius * sqrt(1.0f -vz*vz);
      float radxy2 = radius * sqrt(1.0f -vz2*vz2);
      vertices[vertIdx+0] = radxy * cos(phiv);	// x
      vertices[vertIdx+1] = radxy * sin(phiv);	// y
      vertices[vertIdx+2] = radius * vz;        // z
      vertIdx += 3;
      // Second vertex
      vertices[vertIdx+0] = radxy2 * cos(phiv);	// x
      vertices[vertIdx+1] = radxy2 * sin(phiv);	// y
      vertices[vertIdx+2] = radius * vz2;	// z
      vertIdx += 3;
      // Third vertex
      vertices[vertIdx+0] = radxy2 * cos(phiv2);	// x
      vertices[vertIdx+1] = radxy2 * sin(phiv2);	// y
      vertices[vertIdx+2] = radius * vz2;	        // z
      vertIdx += 3;

      // The second triangle with 3 vertices      
      // First vertex
      vertices[vertIdx+0] = radxy2 * cos(phiv2);	// x
      vertices[vertIdx+1] = radxy2 * sin(phiv2);	// y
      vertices[vertIdx+2] = radius * vz2;	        // z
      vertIdx += 3;
      // Second vertex
      vertices[vertIdx+0] = radxy * cos(phiv2);	// x
      vertices[vertIdx+1] = radxy * sin(phiv2);	// y
      vertices[vertIdx+2] = radius * vz;	        // z
      vertIdx += 3;
      // Third vertex
      vertices[vertIdx+0] = radxy * cos(phiv);	// x
      vertices[vertIdx+1] = radxy * sin(phiv);	// y
      vertices[vertIdx+2] = radius * vz;	        // z
      vertIdx += 3;
    } // for phi
  } // for theta
#if 0
  for (int i = 0; i < cntVertices*3; i+=3) {
    cout << "i=" << i << " " << vertices[i] << " " << vertices[i+1] << " " << vertices[i+2] << "| ";
  }
#endif  
  return cntVertices; // the number of vertices
}

void
CEnvironmentMapGL::init(float radius, int cntTriangles, int renderModeV)
{
  setRenderMode(renderModeV);  

  if (createShaderProgramNoNormals()) {
    cout << "No shader program can be created" << endl;
    exit(-1);
  }

  cntVertices = generateEnv(radius, cntTriangles);  
  
  // Input triangle mesh
  // handles to vertex shader inputs
  int maxattribs;
  glGetIntegerv(GL_MAX_VERTEX_ATTRIBS, &maxattribs);
  //cout << "maxVertexAttribs = " << maxattribs << endl;
  
  // buffer object for the triangles - vertices
  glGenVertexArrays(1, &vaoHandle);
  glBindVertexArray(vaoHandle);
  
  // array of vertex coordinates
  glGenBuffers(1, &vbovertices);
  glBindBuffer(GL_ARRAY_BUFFER, vbovertices);
  int sizeVertices = 3 * sizeof(float) * cntVertices;
  glBufferData(GL_ARRAY_BUFFER, sizeVertices, vertices, GL_STATIC_DRAW);
  CHECK_GL_ERROR();
  //cout << "sizeof(vertices) = " << sizeof(vertices) << endl;
  glBindBuffer(GL_ARRAY_BUFFER, 0);
  //glDisableVertexAttribArray(vertexLoc);
  CHECK_GL_ERROR();
  vertexLoc = glGetAttribLocation(envShaderProgram, "vertex");

  return; // OK
}

void
CEnvironmentMapGL::loadTexture(const char *filename)
{
  // Floating point HDR texture created by reading the file
  CHDRImage hdrimage;
  cout << "Reading file HDR" << endl;
  if (!hdrimage.loadHDR(filename)) {
    cout << "The file " << filename << " cannot be read ... exiting" << endl;
    exit(-2);
  }
  if (texobject == 0) {
    glGenTextures(1, &texobject);
    CHECK_GL_ERROR();
  }
  glBindTexture(GL_TEXTURE_2D, texobject);
  CHECK_GL_ERROR();
  glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_REPEAT);
  glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_REPEAT);
  CHECK_GL_ERROR();
  tex_width = hdrimage.getWidth();
  tex_height = hdrimage.getHeight();
  int cntsize = 3*tex_width*tex_height;
  if (texturedata) delete[] texturedata;
  texturedata = new float[cntsize];
  for (int i = 0; i < cntsize; i++) // copy data
    texturedata[i] = hdrimage.getBuf()[i];
  
  // copy the data to the texture
  glTexImage2D(GL_TEXTURE_2D, 0,  GL_RGBA32F,
	       tex_width, tex_height,
	       0, GL_RGB, GL_FLOAT, &(texturedata[0]));
  CHECK_GL_ERROR();
  glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR);
  glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR);
  CHECK_GL_ERROR();
  glBindTexture(GL_TEXTURE_2D, 0);
  CHECK_GL_ERROR();
  return;
}

void
CEnvironmentMapGL::setTextureData(GLuint envTexID)
{
  texobject = envTexID;
  glBindTexture(GL_TEXTURE_2D, texobject);
  CHECK_GL_ERROR();
  glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_REPEAT);
  glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_REPEAT);
  CHECK_GL_ERROR();
  glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR);
  glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR);
  CHECK_GL_ERROR();
  glBindTexture(GL_TEXTURE_2D, 0);
  CHECK_GL_ERROR();
}

void
CEnvironmentMapGL::setTextureData(float *pointer, int tex_widthV, int tex_heightV)
{
  // buffer for texture created on a PC
  if (texobject == 0) {
    glGenTextures(1, &texobject);
    CHECK_GL_ERROR();
  }
  glBindTexture(GL_TEXTURE_2D, texobject);
  tex_width = tex_widthV;
  tex_height = tex_heightV;
  int cntSize = 4 * tex_width * tex_height;
  if (texturedata) delete[]texturedata;
  texturedata = new float[cntSize];
  assert(texturedata);
  for (int i = 0; i < cntSize; i++)
    texturedata[i]= pointer[i];
#if 0
  // setting some testing value
  for (int i = 0; i < cntSize; i++) {
    texturedata[i] = 1000.f*(float)i/(float)cntSize;
  }
#endif  
  glTexImage2D(   GL_TEXTURE_2D, 
		  0, 
		  GL_RGBA32F,  // GL_RGBA32F, GL_RGB
		  tex_width, tex_height,
		  0, 
		  GL_RGBA, 
		  GL_FLOAT, 
		  &(texturedata[0]));
  CHECK_GL_ERROR();
  glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR);
  glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR);
  CHECK_GL_ERROR();
  glBindTexture(GL_TEXTURE_2D, 0);
  CHECK_GL_ERROR();
  return; // OK
}

void
CEnvironmentMapGL::free()
{
  //glDeleteBuffers(1, &vbonormals);
  if (vbovertices) glDeleteBuffers(1, &vbovertices);
  if (vaoHandle) glDeleteVertexArrays(1, &vaoHandle);
  delete []vertices;
  delete []texturedata;
  //delete []normals;
  if (envShaderProgram) pgr::deleteProgramAndShaders(envShaderProgram);
}

void
CEnvironmentMapGL::draw(const glm::mat4 &PVMmatrix, const glm::mat4 &Mmatrix)
{
  // Binding right program and matrices
  //cout << "envShaderProgram = " << envShaderProgram << endl;
  //glUseProgram(0); //detach old program
  CHECK_GL_ERROR();
  glUseProgram(envShaderProgram);
  CHECK_GL_ERROR();
  // Attribute variables
  //cout << "vertexLoc = " << vertexLoc << endl;
  //cout << "texLoc = " << texLoc << endl;

  glBindVertexArray(vaoHandle);
  CHECK_GL_ERROR();
  glEnableVertexAttribArray(vertexLoc);
  CHECK_GL_ERROR();
  glBindBuffer(GL_ARRAY_BUFFER, vbovertices);
  CHECK_GL_ERROR();
  glVertexAttribPointer(vertexLoc, 3, GL_FLOAT, GL_FALSE, 0, (GLubyte *)NULL);
  CHECK_GL_ERROR();

  // Uniform variables for the vertex shaderr
  GLint PVMmatLoc = glGetUniformLocation(envShaderProgram, "PVMmatrix");
  glUniformMatrix4fv(PVMmatLoc, 1, GL_FALSE, glm::value_ptr(PVMmatrix)); // binding matrix to shader
  CHECK_GL_ERROR();
  // Uniform variables for the texture shader
  GLint samplerID = glGetUniformLocation(envShaderProgram, "texsampler");
  GLint modelMatID = glGetUniformLocation(envShaderProgram, "Mmatrix");
  glUniformMatrix4fv(modelMatID, 1, GL_FALSE, glm::value_ptr(Mmatrix)); // binding matrix to shader
  CHECK_GL_ERROR();
  glUniform1i(samplerID, 0); // was glUniform1i(  glUniform1f(
  CHECK_GL_ERROR();
  glActiveTexture(GL_TEXTURE0 + 0);
  CHECK_GL_ERROR();

  // Binding vertex array object
  glBindVertexArray(vaoHandle);
  glBindTexture(GL_TEXTURE_2D, texobject);
  glDrawArrays(renderMode, 0, cntVertices);
  CHECK_GL_ERROR();
  glBindVertexArray(0); // detach
  CHECK_GL_ERROR();
  glBindTexture(GL_TEXTURE_2D, 0);
  CHECK_GL_ERROR();
  glUseProgram(0);
  CHECK_GL_ERROR();
  return;
}

