// ================================================================
// $Id: mapprox.h,v 1.2 2005/11/02 12:58:44 xkrivanj Exp $
//
// maprox.h
//     Basic mathematical definitions
//
// Class: CLimits
//
// REPLACEMENT_STRING
//
// Initial coding by Vlastimil Havran, December 2004.

#ifndef __MAPPROX_H__
#define __MAPPROX_H__

#include "configh.h"
#include "basmath.h"

// forward declarations

// The fast computation of arctangent .. the maximal error is less
// than 4.1 degrees, according to Graphics GEMSII, 1991, pages 389--391
// Ron Capelli: "Fast approximation to the arctangent"
inline float
atan22(const float &y)
{
  const float x = 1.0;
  const float c = (float)(M_PI * 0.25);
  
  if (y < 0.0) {
    if (y < -1.0)
      return c * (-2.0 + x / y); // for angle in <-PI/2, -PI/4)
    else
      return c * (y / x); // for angle in <-PI/4 , 0>
  }
  else {
    if (y > 1.0)
      return c * (2.0 - x / y); // for angle in <PI/4, PI/2>
    else
      return c * (y / x); // for angle in <0, PI/2>
  }
}

// -----------------------------------------------------------
// This is the implementation of atan2 (approximation), found
// on the web: bassw@hotmail.com, 2004-10-21.
// The accuracy is around 2*M_PI/1000 (= 0.005 radians).
inline float
atan2approx(const float &y, const float &x)
{
  float f, offset = 0;
  // Based on quadrants
  if (y > x) {
    if (y > -x) {// north
      f = -x / y;
      offset = M_PI_2;
    }
    else { // west
      f = y / x;
      if (y > 0.0) // or use >= determines if (ata2n(0,01) is -pi or pi
	offset = M_PI;
      else
	offset = -M_PI;
    }
  }
  else {
    if (y > -x) { // east
      f = y / x;
      offset = 0.0;   
    }
    else {
      if (y == 0.0) {
	f = 0.0;
        offset = -M_PI_2;
      }
      else {
	f = -x/y;
	offset = -M_PI_2;
      }
    }
  }

  // Compute the approximation by cubic function
  return offset + f * (0.97179803007662069f - 0.19065470515032634 * f * f);  
}

#endif // __MAPPROX_H__
